#include "Python.h"
#include <unicode/convert.h>
#include <unicode/bidi.h>
#include <unicode/unistr.h>
#include <stdio.h>
#include <assert.h>

#define isheb(c) (((((unsigned char) c)>=224) && (((unsigned char) c)<=250)) ? 1 : 0)

#define LINESIZE 8192
#define checkfailure(ec) if (! U_SUCCESS(ec)) { return 0; } 

char *_log2vis(char *str, int str_len, UBiDiLevel paraLevel, char *inencoding, char *outencoding)
{
    UErrorCode errorcode = U_ZERO_ERROR;
    UnicodeConverterCPP inconv(inencoding, errorcode);
    checkfailure(errorcode);
    
    UnicodeConverterCPP outconv(outencoding, errorcode);
    checkfailure(errorcode);
    
    UnicodeString ustr, resultString;
    BiDi bidi(LINESIZE, LINESIZE, errorcode);
    checkfailure(errorcode);
    
    char *result;
    UChar uresult[LINESIZE];
    UTextOffset uresultlen;
    
    /* Convert from input encoding to Unicode */
    inconv.toUnicodeString(ustr, str, str_len, errorcode);
    checkfailure(errorcode);
    
    /* Convert from Logical to Visual */
    bidi.setPara(ustr.getUChars(), ustr.size(), paraLevel, (UBiDiLevel *)NULL, errorcode);
    checkfailure(errorcode);
    uresultlen = bidi.writeReordered(uresult, LINESIZE, UBIDI_REMOVE_BIDI_CONTROLS, errorcode);
    checkfailure(errorcode);
    
    /* Convert from Unicode to output encoding */
    result = (char *) malloc(sizeof(char) * (uresultlen + 1));
    resultString = UnicodeString(uresult, uresultlen);
    outconv.fromUnicodeString(result, uresultlen, resultString, errorcode);
    if (! U_SUCCESS(errorcode)) {
        free(result);
        return 0;
    }
    
    result[uresultlen] = 0;
    return result;
}


char log2vis__doc__[] = "log2vis(line, direction, inencoding='iso-8859-8', \
outencoding='iso-8859-8'): Convert a line from logical to visual text.";

static PyObject *log2vis_log2vis(PyObject *self, PyObject *args, PyObject *kwargs) 
{
    char *kwlist[] = {"inencoding", "outencoding", NULL};
    char *str, *result;
    char *inencoding = "iso-8859-8", *outencoding = "iso-8859-8";
    int str_len, paraLevel;
    PyObject *pyResult;
    
    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s#i|ss", kwlist, 
                          &str, &str_len, &paraLevel, &inencoding, &outencoding)) {
        return NULL;
    }
    
    /*
    if isheb(str[0]) {
        paraLevel = 1;
    } else {
        paraLevel = 0;
    } */
    
    /* Don't bother with algorithm for empty/over-long strings (core dumps) */
    if ((str_len > 0) && (str_len < LINESIZE)) {
        result = _log2vis(str, str_len, paraLevel, inencoding, outencoding);
    } else {
        result = (char *)malloc(sizeof(char));
        result[0] = '\x0';
    }
    
    if (! result == 0) {
        pyResult = Py_BuildValue("s", result);
        free(result);
        return pyResult;
    } else {
        return NULL;
    }
}


static PyMethodDef log2visMethods[] = {
    {"log2vis", (PyCFunction)log2vis_log2vis, METH_VARARGS|METH_KEYWORDS, log2vis__doc__},      
    {NULL,      NULL}        /* Sentinel */
};

extern "C" {
    void init_log2vis(void)
    {
        (void) Py_InitModule("_log2vis", log2visMethods);
    }
}
